import java.util.Random;
import java.awt.Color;

/*************************************************************************
 *  Compilation:  javac Maze.java
 *  Execution:    java Maze.java N
 *  Dependecies:  StdDraw.java
 *
 *  Generates a perfect N-by-N maze using depth-first search with a stack.
 *
 *  % java Maze 62
 *
 *  % java Maze 61
 *
 *  Note: this program generalizes nicely to finding a random tree
 *        in a graph.
 *
 *  Adapted from http://algs4.cs.princeton.edu/41undirected/Maze.java.html
 *  (by Kevin Wayne at Princeton University)
 *
 *************************************************************************/

public class Maze {

    // solve the maze using depth-first search
    private void findCenter(int x, int y) {     
        if (isTarget(x, y)) {
            celebrate();
        }   

        if (!hasCrumb(x, y)) {
            leaveCrumb(x, y);

            if (!wallExists(x, y, "North")) {
                findCenter(x, y+1);
            }

            if (!wallExists(x, y, "West")) {
                findCenter(x-1, y);
            }

            if (!wallExists(x, y, "South")) {
                findCenter(x, y-1);
            }

            if (!wallExists(x, y, "East")) {
                findCenter(x + 1, y);
            }
        }

    }

    // solve the maze using depth-first search
    private void findCenterStackBased(int x, int y) {        
        Stack<Integer> s = new Stack<Integer>(); 
        s.push(x);
        s.push(y);
        while (!s.isEmpty()) {
            System.out.println(s);

            y = s.pop();
            x = s.pop();

            if (isTarget(x, y)) {
                celebrate();
            }   

            if (!hasCrumb(x, y)) {
                leaveCrumb(x, y);

                if (!wallExists(x, y, "North")) {
                    s.push(x);
                    s.push(y + 1);
                }

                if (!wallExists(x, y, "West")) {
                    s.push(x-1);
                    s.push(y);
                }

                if (!wallExists(x, y, "South")) {
                    s.push(x);
                    s.push(y-1);
                }

                if (!wallExists(x, y, "East")) {
                    s.push(x+1);
                    s.push(y);
                }
            }            
        }
    }

    /* draws a dot in position x, y of color c */
    private void leaveCrumb(int x, int y) {
        draw(x, y, StdDraw.BLUE);                
        crumb[x][y] = true;
    }

    /* draws a dot in position x, y of color c */
    private boolean hasCrumb(int x, int y) {
        return crumb[x][y];
    }


    // returns true if wall exists
    private boolean wallExists(int x, int y, String s) {
        int tx = targetX(x, s);
        int ty = targetY(y, s);
        boolean ooBounds = (tx == 0 || ty == 0 || tx == N+1 || ty == N+1);

        if (ooBounds)
            return true;

        if (s.equals("North"))
            return north[x][y];

        if (s.equals("East"))
            return east[x][y];

        if (s.equals("South"))
            return south[x][y];

        if (s.equals("West"))
            return west[x][y];

        return true;
    }










    private int N;                 // dimension of maze
    private boolean[][] north;     // is there a wall to north of cell i, j
    private boolean[][] east;
    private boolean[][] south;
    private boolean[][] west;
    private boolean[][] crumb;
    private static int moveCount;
    private boolean done = false;
    private static Random rgen;
    private static final int DRAW_DELAY_MS = 300;
    private static final boolean USE_STACK_BASED_SOLUTION = false;


    public Maze(int N, int rseed) {
        this.N = N;
        StdDraw.setXscale(0, N+2);
        StdDraw.setYscale(0, N+2);
        moveCount = 0;
        rgen = new Random(rseed);
        init();
        generate();
    }

    private void init() {
        // initialize border cells as already crumb
        crumb = new boolean[N+2][N+2];
        for (int x = 0; x < N+2; x++) crumb[x][0] = crumb[x][N+1] = true;
        for (int y = 0; y < N+2; y++) crumb[0][y] = crumb[N+1][y] = true;


        // initialze all walls as present
        north = new boolean[N+2][N+2];
        east  = new boolean[N+2][N+2];
        south = new boolean[N+2][N+2];
        west  = new boolean[N+2][N+2];
        for (int x = 0; x < N+2; x++)
            for (int y = 0; y < N+2; y++)
                north[x][y] = east[x][y] = south[x][y] = west[x][y] = true;
    }


    // generate the maze
    private void generate(int x, int y) {
        crumb[x][y] = true;
        // while there is an uncrumb neighbor
        while (!crumb[x][y+1] || !crumb[x+1][y] || !crumb[x][y-1] || !crumb[x-1][y]) {

            // pick random neighbor (could use Knuth's trick instead)
            while (true) {
                
                double r = rgen.nextDouble();

                if (r < 0.25 && !crumb[x][y+1]) {
                    north[x][y] = south[x][y+1] = false;
                    generate(x, y + 1);
                    break;
                }
                else if (r >= 0.25 && r < 0.50 && !crumb[x+1][y]) {
                    east[x][y] = west[x+1][y] = false;
                    generate(x+1, y);
                    break;
                }
                else if (r >= 0.5 && r < 0.75 && !crumb[x][y-1]) {
                    south[x][y] = north[x][y-1] = false;
                    generate(x, y-1);
                    break;
                }
                else if (r >= 0.75 && r < 1.00 && !crumb[x-1][y]) {
                    west[x][y] = east[x-1][y] = false;
                    generate(x-1, y);
                    break;
                }
            }
        }
    }

    // generate the maze starting from lower left
    private void generate() {
        generate(1, 1);    
    }


    // returns true if current position is target
    private boolean isTarget(int x, int y) {
        return x == N/2 && y == N/2;
    }

    // Returns x coordinate of target given source location
    // and direction
    private int targetX(int x, String s) {
        if (s.equals("West"))
            return x - 1;
        if (s.equals("East"))
            return x + 1;
        return x;
    }

    // Returns y coordinate of target given source location
    // and direction
    private int targetY(int y, String s) {
        if (s.equals("South"))
            return y - 1;
        if (s.equals("North"))
            return y + 1;
        return y;
    }

    /* draws a dot in position x, y of color c */
    private static void draw(int x, int y, Color c) {
        StdDraw.setPenColor(c);
        StdDraw.filledCircle(x + 0.5, y + 0.5, 0.25);
        StdDraw.setPenColor(StdDraw.WHITE);
        StdDraw.text(x + 0.5, y + 0.5, Integer.toString(moveCount));        
        StdDraw.show(DRAW_DELAY_MS);        
        moveCount += 1;
    }

    private void celebrate() {
        
        draw(N/2,  N/2, StdDraw.BLUE);
        StdDraw.setPenColor(StdDraw.MAGENTA);        
        StdDraw.text(5, 5, "yay.");
        StdDraw.show();
        try {
            Thread.sleep(400000);        
        } catch(InterruptedException e) {
            System.out.println("This happened: " + e);
        }
    }

    // solve the maze starting from the start state
    public void solve() {
        for (int x = 1; x <= N; x++)
            for (int y = 1; y <= N; y++)
                crumb[x][y] = false;
        done = false;
        if (USE_STACK_BASED_SOLUTION)
            findCenterStackBased(N - 1, N - 1);
        else
            findCenter(N-1, N-1);
    }

    // draw the maze
    public void draw() {
        StdDraw.setPenColor(StdDraw.RED);
        StdDraw.filledCircle(N/2.0 + 0.5, N/2.0 + 0.5, 0.375);
        StdDraw.filledCircle(N-0.5, N-0.5, 0.375);

        StdDraw.setPenColor(StdDraw.BLACK);
        for (int x = 1; x <= N; x++) {
            for (int y = 1; y <= N; y++) {
                if (south[x][y]) StdDraw.line(x, y, x + 1, y);
                if (north[x][y]) StdDraw.line(x, y + 1, x + 1, y + 1);
                if (west[x][y])  StdDraw.line(x, y, x, y + 1);
                if (east[x][y])  StdDraw.line(x + 1, y, x + 1, y + 1);
            }
        }
        StdDraw.show(1000);
    }



    // a test client
    public static void main(String[] args) {
        int N = Integer.parseInt(args[0]);
        int rseed = Integer.parseInt(args[1]);
        Maze maze = new Maze(N, rseed);
        StdDraw.show(0);
        maze.draw();
        maze.solve();
    }

}

